/* ml_exthdr.c - supporting RFC-1342 (Non-ASCII Mail Headers) */
/*				by takada@seraph.ntt.jp       */

#ifdef JAPAN

#include "../h/mh.h"
#include <stdio.h>
#include <ctype.h>

#define MAXWIDTH 75

static char b64_to_alpha[] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static unsigned char alpha_to_b64[256] = {
     0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
     0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
     0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0,62,  0, 0, 0,63,
    52,53,54,55, 56,57,58,59, 60,61, 0, 0,  0, 0, 0, 0,
     0, 0, 1, 2,  3, 4, 5, 6,  7, 8, 9,10, 11,12,13,14,
    15,16,17,18, 19,20,21,22, 23,24,25, 0,  0, 0, 0, 0,
     0,26,27,28, 29,30,31,32, 33,34,35,36, 37,38,39,40,
    41,42,43,44, 45,46,47,48, 49,50,51, 0,  0, 0, 0, 0,
     0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
     0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
     0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
     0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
     0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
     0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
     0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
     0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
};

static int  ml_to_mmh();
static int  mmh_to_ml();
static int  bin_to_b64();
static int  b64_to_bin();


/*
 * Encode multilingual string into RFC-1342 format
 */
char *
exthdr_encode(input, output, width)
     char *input, *output;
{
    (void) ml_to_mmh(input, output, width);
    return(output);
}

#define ASCII 		0
#define JISX0208	1

#define DSGNT_JISX0208() {\
    kanji_pos = JISX0208;\
}

#define DSGNT_ASCII() {\
    *bp++ = '\033'; *bp++ = '('; *bp++ = 'B'; *bp = 0;\
    strcpy(op, "=?ISO-2022-JP?B?"); op += 16; width += 16;\
    x = bin_to_b64(buf, op); op += x; width += x;\
    strcpy(op, "?="); op += 2; width += 2;\
    bp = buf; *bp++ = '\033'; *bp++ = '$'; *bp++ = 'B';\
    kanji_pos = ASCII;\
}

static int
ml_to_mmh(ibuf, obuf, width)
     char *ibuf, *obuf;
     int width;
{
    int kanji_pos, x;
    char *ip = ibuf;
    char *op = obuf;
    char *bp, buf[BUFSIZ];
    unsigned char c1, c2;

    bp = buf; *bp++ = '\033'; *bp++ = '$'; *bp++ = 'B';
    kanji_pos = ASCII;
    
    while (c1 = *ip++) {
	if (ml_ismlchar(c1)) {
	    if (width + ((bp-buf+2+3+2)/3)*4 + 16 + 2 >= MAXWIDTH) {
		if (kanji_pos == JISX0208) {
		    DSGNT_ASCII();
		    *op++ = '\n'; *op++ = '\t'; width = 8;
		    DSGNT_JISX0208();
		} else {
		    *op++ = '\n'; *op++ = '\t'; width = 8;
		}
	    }
	} else {
	    if ((width >= MAXWIDTH) && isspace(c1) && c1 != '\n') {
		*op++ = '\n'; *op++ = '\t'; width = 8;
	    }
	}
	if (ml_ismlchar(c1)) {
	    if (c2 = *ip++) {
		if (ml_ismlchar(c2)) {
		    if (kanji_pos == ASCII) {
			DSGNT_JISX0208();
		    }
		    *bp++ = c1 & 0x7f; *bp++ = c2 & 0x7f;
		} else {
		    if (kanji_pos == JISX0208) {
			DSGNT_ASCII(); *op++ = ' '; width++;
		    }
		    *op++ = ' '; *op++ = c2; width += 2;
		}
	    } else {
		if (kanji_pos == JISX0208) {
		    DSGNT_ASCII(); *op++ = ' '; width++;
		}
		*op++ = c1; width++;
		break;
	    }
	} else if (c1 == '\n') {
	    if (kanji_pos == JISX0208) {
		DSGNT_ASCII();
	    }
	    *op++ = c1; width = 0;
	} else {
	    if (kanji_pos == JISX0208) {
		DSGNT_ASCII(); *op++ = ' '; width++;
	    }
	    *op++ = c1; width++;
	}
    }
    switch (kanji_pos) {
      case JISX0208:
	DSGNT_ASCII();
	*op++ = ' '; width++;
	break;
      case ASCII:
	DSGNT_JISX0208();
	break;
    }
    *op = 0;
    return(op - obuf);
}

static int
bin_to_b64(in, out)
     char *in, *out;
{
    char *oldout = out;
    char c1, c2, c3;

    while (c1 = *in++) {
	*out++ = b64_to_alpha[c1 >> 2];
	c2 = *in++;
	*out++ = b64_to_alpha[((c1 & 0x03) << 4) | ((c2 & 0xf0) >> 4)];
	if (c2) {
	    c3 = *in++;
	    *out++ = b64_to_alpha[((c2 & 0x0f) << 2) | ((c3 & 0xc0) >> 6)];
	    if (c3) {
		*out++ = b64_to_alpha[c3 & 0x3f];
	    } else {
		*out++ = '=';
		break;
	    }
	} else {
	    *out++ = '=';
	    *out++ = '=';
	    break;
	}
    }
    *out = 0;
    return(out - oldout);
}


/*
 * Decode multilingual string from RFC-1342 format
 */
char *
exthdr_decode(input, output)
     char *input, *output;
{
    char *cp = input;
    char *op = output;
    int  in, out;

    while (*cp) {
	if (uprf(cp, "=?") &&
	    ((in = stringdex("?=", cp)) >= 0) &&
	    ((out = mmh_to_ml(cp, cp+in+1, op)) >= 0)) {
	    cp += in + 2; op += out;
	    if (!*cp) {
		break;
	    } else if (*cp == '\n' && *(cp+1)) {
		while (*++cp && isspace(*cp)) ;
	    } else if (isspace(*cp) && *(cp+1)) {
		cp++;
	    }
	} else {
	    *op++ = *cp++;
	}
    }
    *op = NULL;
    return(output);
}

static int
mmh_to_ml(bp, ep, op)
     char *bp, *ep, *op;
{
    char *ip, *cp, buf[BUFSIZ];
    int  result;

    ip = bp;
    if ((ip > ep) || strncmp(ip, "=?", 2)) return(-1);
    ip += 2;
    if ((ip > ep) || ((cp = index(ip, '?')) == NULL)) return(-1);
    strncpy(buf, ip, cp-ip); buf[cp-ip] = 0;
    if (!uleq(buf, "iso-2022-jp") && !uleq(buf, "x-iso-2022-jp")) return(-1);
    ip = cp + 1;
    if ((ip > ep) || ((*ip != 'B') && (*ip != 'b'))) return(-1);
    ip++;
    if ((ip > ep) || (*ip != '?')) return(-1);
    ip++;
    if ((ip > ep) || ((cp = index(ip, '?')) == NULL)) return(-1);
    strncpy(buf, ip, cp-ip); buf[cp-ip] = 0;
    ip = cp + 1;
    if ((ip != ep) || (*ip != '=')) return(-1);
    (void) b64_to_bin(buf, op);
    (void) ml_conv(op);
    return(strlen(op));
}

static int
b64_to_bin(in, out)
     char *in, *out;
{
    char *oldout = out;
    unsigned char c1, c2, c3, c4;

    while (c1 = *in++) {
	if ((c2 = *in++) && (c3 = *in++) && (c4 = *in++)) {
	    if ((c1 == '=') || (c2 == '=')) {
		break;
	    } else {
		c1 = alpha_to_b64[c1];
		c2 = alpha_to_b64[c2];
	    }
	    *out++ = (c1 << 2) | ((c2 & 0x30) >> 4);
	    if (c3 == '=') {
		break;
	    } else {
		c3 = alpha_to_b64[c3];
	    }
	    *out++ = ((c2 & 0x0f) << 4) | ((c3 & 0x3c) >> 2);
	    if (c4 == '=') {
		break;
	    } else {
		c4 = alpha_to_b64[c4];
	    }
	    *out++ = ((c3 & 0x03) << 6) | c4 ;
	} else {
	    break;
	}
    }
    *out = 0;
    return(out - oldout);
}	    

#endif /* JAPAN */
